# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import urlparse

from lmi.shell.compat import *

from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMIUtil import lmi_parse_uri
from lmi.shell.LMIUtil import lmi_raise_or_dump_exception
from lmi.shell.LMIUtil import lmi_get_use_exceptions
from lmi.shell.LMIUtil import lmi_set_use_exceptions
from lmi.shell.LMIUtil import lmi_instance_to_path
from lmi.shell.LMIUtil import lmi_is_localhost

from lmi.shell.LMIDecorators import lmi_process_cim_exceptions
from lmi.shell.LMIDecorators import lmi_process_cim_exceptions_rval

from lmi.shell.LMIExceptions import CIMError
from lmi.shell.LMIExceptions import ConnectionError
from lmi.shell.LMIExceptions import LMIFilterError


def filter_kwargs(**kwargs):
    return {k: v for k, v in kwargs.iteritems() if v is not None}


class LMICIMXMLClient(object):
    """
    CIM-XML client.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be
        performed
    :param string password: user's password
    :param bool verify_server_cert: indicates, whether a server side
        certificate needs to be verified, if SSL used; default value is True
    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    """
    QUERY_LANG_CQL = "DMTF:CQL"
    QUERY_LANG_WQL = "WQL"

    # Unix sockets used by TOG-Pegasus, SFCB or OpenWBEM.
    PEGASUS_SOCKET = "/var/run/tog-pegasus/cimxml.socket"

    def __init__(self, uri, username="", password="",
                 verify_server_cert=True,
                 key_file=None, cert_file=None):
        self._uri = uri
        if not self._uri.startswith("http://") and \
                not self._uri.startswith("https://"):
            self._uri = "https://" + uri

        # Parse CIMOM URI
        _, hostname, _, path, username_, password_ = lmi_parse_uri(self._uri)
        if username_ is not None:
            username = username_
            password = password_

        creds = (username, password)
        x509 = {
            "cert_file": cert_file,
            "key_file": key_file
        }

        # If we are connecting to localhost without username or certificate
        # set, we will use the local UNIX socket for higher performance (and
        # skip authentication).
        connect_locally = lmi_is_localhost(uri) and \
            path != "/wsman" and \
            not username and not password and \
            not cert_file and not key_file

        if HAVE_LMIWBEM:
            # LMIWBEM
            self._cliconn = wbem.WBEMConnection(
                self._uri,
                creds=creds,
                x509=x509,
                no_verification=not verify_server_cert,
                connect_locally=connect_locally)
        else:
            # PyWBEM
            if connect_locally:
                self._cliconn = wbem.PegasusUDSConnection()
            else:
                self._cliconn = wbem.WBEMConnection(
                    self._uri,
                    creds=creds,
                    x509=x509,
                    no_verification=not verify_server_cert)

                # LMIWBEM defines hostname attribute, so we define it here,
                # as well.
                self._cliconn.hostname = hostname

    def __del__(self):
        """
        Disconnects and frees :py:class:`WBEMConnection` object.

        **NOTE:** Applicable only wbem :py:mod:`lmiwbem` is used.
        """
        if HAVE_LMIWBEM:
            self.disconnect()
            del self._cliconn

    @lmi_process_cim_exceptions(False)
    def connect(self):
        """
        Connects to CIMOM.

        **NOTE:** Applicable only wbem :py:mod:`lmiwbem` is used.
        """
        if HAVE_LMIWBEM:
            self._cliconn.connect()
        return LMIReturnValue(rval=True)

    def disconnect(self):
        """
        Disconnects from CIMOM.

        **NOTE:** Applicable only wbem :py:mod:`lmiwbem` is used.
        """
        if HAVE_LMIWBEM:
            self._cliconn.disconnect()

    def dummy(self):
        """
        Sends a "dummy" request to verify credentials.

        :returns: :py:class:`.LMIReturnValue` with rval set to True, if
            provided credentials are OK; False otherwise. If LMIShell uses
            exceptions, :py:exc:`.CIMError` or :py:exc:`.ConnectionError` will
            be raised.
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        try:
            use_exceptions = lmi_get_use_exceptions()
            lmi_set_use_exceptions(True)
            try:
                # Perform a dummy request
                self.get_class("SomeNonExistingClass")
            except:
                raise
            finally:
                lmi_set_use_exceptions(use_exceptions)
        except CIMError, e:
            if e.args[0] == wbem.CIM_ERR_NOT_FOUND:
                return LMIReturnValue(rval=True)
            lmi_raise_or_dump_exception(e)
            errorstr = e.args[1]
        except ConnectionError, e:
            lmi_raise_or_dump_exception(e)
            errorstr = e.args[1]

        return LMIReturnValue(rval=False, errorstr=errorstr)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_process_cim_exceptions()
    def get_instance_names(self, classname, namespace=None, inst_filter=None,
                           limit=-1, **kwargs):
        """
        Returns a list of :py:class:`wbem.CIMInstanceName` objects.

        :param string classname: class name
        :param string namespace: namespace name, where the instance names live
        :param dictionary inst_filter: dictionary containing filter values. The
            key corresponds to the primary key of the
            :py:class:`wbem.CIMInstanceName`; value contains the filtering
            value.
        :param int limit: unused
        :param dictionary kwargs: supported keyword arguments (these are
            **deprecated**)

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** (*string*) -- filtering value, see above

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` contains a
            list of :py:class:`wbem.CIMInstanceName` objects, if no error
            occurs; otherwise ``rval`` is set to None and ``errorstr`` contains
            appropriate error string
        :raises: :py:exc:`.LMIFilterError`, :py:exc:`.CIMError`,
            :py:exc:`.ConnectionError`
        """
        filter_value = ""
        filter_key = "Name"
        if inst_filter is None:
            inst_filter = {}
        if "key" in kwargs:
            filter_key = kwargs["key"]
            kwargs.pop("key")
        elif "Key" in kwargs:
            filter_key = kwargs["Key"]
            kwargs.pop("Key")
        if "value" in kwargs:
            filter_value = kwargs["value"]
            kwargs.pop("value")
        if "Value" in kwargs:
            filter_value = kwargs["Value"]
            kwargs.pop("Value")
        if filter_value:
            inst_filter[filter_key] = filter_value
        try:
            inst_name_list = self._cliconn.EnumerateInstanceNames(
                classname, namespace)
            if inst_filter:
                inst_name_list_filtered = []
                for inst_name in inst_name_list:
                    append = True
                    for filter_key, filter_value in inst_filter.iteritems():
                        if inst_name[filter_key] != filter_value:
                            append = False
                            break
                    if append:
                        inst_name_list_filtered.append(inst_name)
                inst_name_list = inst_name_list_filtered
        except KeyError, e:
            errorstr = "Can not filter by '%s'" % filter_key
            lmi_raise_or_dump_exception(LMIFilterError(errorstr))
            return LMIReturnValue(rval=None, errorstr=errorstr)
        return LMIReturnValue(rval=inst_name_list)

    @lmi_process_cim_exceptions()
    def get_instance(self, instance, LocalOnly=True, IncludeQualifiers=False,
                     IncludeClassOrigin=False, PropertyList=None):
        """
        Returns a :py:class:`wbem.CIMInstance` object.

        :param path: path of the object, which is about to be retrieved. The
            object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`wbem.CIMInstance`
            * :py:class:`.LMIInstanceName`
            * :py:class:`.LMIInstance`

        :param bool LocalOnly: indicates if to include the only elements
            (properties, methods, references) overridden or defined in the
            class
        :param bool IncludeQualifiers: indicates, if all Qualifiers for the
            class and its elements shall be included in the response
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in the
            returned class
        :param list PropertyList: if present and not None, the members of the
            list define one or more property names. The returned class shall
            not include elements for properties missing from this list. Note
            that if LocalOnly is specified as True, it acts as an additional
            filter on the set of properties returned. For example, if property
            A is included in the PropertyList but LocalOnly is set to True and
            A is not local to the requested class, it is not included in the
            response. If the PropertyList input parameter is an empty list, no
            properties are included in the response. If the PropertyList input
            parameter is None, no additional filtering is defined.
        :returns: :py:class:`.LMIReturnValue` object, where ``rval`` is set to
            :py:class:`wbem.CIMInstance` object, if no error occurs; otherwise
            ``rval`` is set to None and ``errorstr`` is set to corresponding
            error string.
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        return LMIReturnValue(
            rval=self._cliconn.GetInstance(
                lmi_instance_to_path(instance),
                **filter_kwargs(
                    LocalOnly=LocalOnly,
                    IncludeQualifiers=IncludeQualifiers,
                    IncludeClassOrigin=IncludeClassOrigin,
                    PropertyList=PropertyList)))

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_process_cim_exceptions()
    def get_instances(self, classname, namespace=None, inst_filter=None,
                      client_filtering=False, limit=-1, **kwargs):
        """
        Returns a list of :py:class:`wbem.CIMInstance` objects.

        :param string classname: class name
        :param string namespace: namespace, where the instances live
        :param dictionary inst_filter: dictionary containing filter values. The
            key corresponds to the primary key of the
            :py:class:`wbem.CIMInstanceName`; value contains the filtering
            value.
        :param bool client_filtering: if True, client-side filtering will be
            performed, otherwise the filtering will be done by a CIMOM. Default
            value is False.
        :param int limit: unused
        :param dictionary kwargs: supported keyword arguments (these are
            **deprecated**)

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** (*string*) -- filtering value, see above

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of :py:class:`wbem.CIMIntance` objects, if no error occurs;
            otherwise ``rval`` is set to None and ``errorstr`` is set to
            corresponding error string.
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        filter_value = ""
        filter_key = "Name"
        if inst_filter is None:
            inst_filter = {}
        if "key" in kwargs:
            filter_key = kwargs["key"]
            kwargs.pop("key")
        elif "Key" in kwargs:
            filter_key = kwargs["Key"]
            kwargs.pop("Key")
        if "value" in kwargs:
            filter_value = kwargs["value"]
            kwargs.pop("value")
        if "Value" in kwargs:
            filter_value = kwargs["Value"]
            kwargs.pop("Value")
        if filter_value:
            inst_filter[filter_key] = filter_value

        if not client_filtering:
            query = "select * from %s" % classname
            if inst_filter:
                more = False
                query += " where"
                for filter_key, filter_value in inst_filter.iteritems():
                    if more:
                        query += " and"
                    quotes = isinstance(filter_value, basestring)
                    query += " %s =" % filter_key
                    if quotes:
                        query += " \"%s\"" % filter_value
                    else:
                        query += " %s" % filter_value
                    more = True
            inst_list, _, errorstr = self.exec_query(
                LMICIMXMLClient.QUERY_LANG_WQL, query, namespace)
            if inst_list is None:
                return LMIReturnValue(rval=None, errorstr=errorstr)
            return LMIReturnValue(rval=inst_list)

        # Client-side filtering - this is not a pretty solution, but it needs
        # to be present due to TOG-Pegasus, which does not raise an exception,
        # if an error occurs while performing CQL/WQL query.
        inst_list = self._cliconn.EnumerateInstances(
            classname,
            namespace,
            LocalOnly=False,
            DeepInheritance=True,
            IncludeQualifiers=False)

        if inst_list and inst_filter:
            inst_list_filtered = []
            for inst in inst_list:
                for filter_key, filter_value in inst_filter.iteritems():
                    if filter_key not in inst.properties or \
                            inst.properties[filter_key].value != filter_value:
                        break
                else:
                    inst_list_filtered.append(inst)
            inst_list = inst_list_filtered
        return LMIReturnValue(rval=inst_list)

    @lmi_process_cim_exceptions()
    def get_class_names(self, namespace=None, ClassName=None,
                        DeepInheritance=False):
        """
        Returns a list of class names.

        :param string namespace: namespace, from which the class names list
            should be retrieved; if None, default namespace will be used
            (**NOTE:** see :py:mod:`wbem`)
        :param string ClassName: defines the class that is the basis for the
            enumeration.  If the ClassName input parameter is absent, this
            implies that the names of all classes.
        :param bool DeepInheritance: if not present, of False, only the names
            of immediate child subclasses are returned, otherwise the names of
            all subclasses of the specified class should be returned.
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of strings containing class names, if no error occurs;
            otherwise ``rval`` is set to None and ``errorstr``
            contains an appropriate error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        return LMIReturnValue(
            rval=self._cliconn.EnumerateClassNames(
                namespace,
                **filter_kwargs(
                    ClassName=ClassName,
                    DeepInheritance=DeepInheritance)))

    @lmi_process_cim_exceptions()
    def get_class(self, classname, namespace=None, LocalOnly=True,
                  IncludeQualifiers=True, IncludeClassOrigin=False,
                  PropertyList=None):
        """
        Returns a :py:class:`wbem.CIMClass` object.

        :param string classname: class name
        :param string namespace: -- namespace name, from which the
            :py:class:`wbem.CIMClass` should be retrieved; if None, default
            namespace will be used (**NOTE:** see :py:mod:`wbem`)
        :param bool LocalOnly: indicates, if only local members should be
            present in the returned :py:class:`wbem.CIMClass`; any CIM elements
            (properties, methods, and qualifiers), except those added or
            overridden in the class as specified in the classname input
            parameter, shall not be included in the returned class.
        :param bool IncludeQualifiers: indicates, if qualifiers for the class
            (including qualifiers on the class and on any returned properties,
            methods, or method parameters) shall be included in the response.
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in the
            returned class.
        :param list PropertyList: if present and not None, the members of the
            list define one or more property names. The returned class shall
            not include elements for properties missing from this list. Note
            that if LocalOnly is specified as True, it acts as an additional
            filter on the set of properties returned. For example, if property
            A is included in the PropertyList but LocalOnly is set to True and
            A is not local to the requested class, it is not included in the
            response. If the PropertyList input parameter is an empty list, no
            properties are included in the response. If the PropertyList input
            parameter is None, no additional filtering is defined.
        :returns: :py:class:`.LMIReturnValue` object with rval set to
            :py:class:`wbem.CIMClass`, if no error occurs; otherwise rval is
            set to None and errorstr to appropriate error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        return LMIReturnValue(
            rval=self._cliconn.GetClass(
                classname,
                namespace,
                **filter_kwargs(LocalOnly=LocalOnly,
                    IncludeQualifiers=IncludeQualifiers,
                    IncludeClassOrigin=IncludeClassOrigin,
                    PropertyList=PropertyList)))

    def get_superclass(self, classname, namespace=None):
        """
        Returns a superclass to given class.

        :param string classname: class name
        :param string namespace: namespace name
        :returns: superclass to given class, if such superclass exists,
            None otherwise
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        minimal_class, rparams, errorstr = LMICIMXMLClient.get_class(
            self, classname, namespace, LocalOnly=True,
            IncludeQualifiers=False, PropertyList=[])
        if not minimal_class:
            return LMIReturnValue(
                rval=None, rparams=rparams, errorstr=errorstr)
        return LMIReturnValue(rval=minimal_class.superclass)

    @lmi_process_cim_exceptions(-1)
    def call_method(self, instance, method, **params):
        """
        Executes a method within a given instance.

        :param instance: object, on which the method will be executed. The
            object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string method: string containing a method name
        :param dictionary params: parameters passed to the method call
        :returns: :py:class:`.LMIReturnValue` object with rval set to return
            value of the method call, rparams set to returned parameters from
            the method call, if no error occurs; otherwise rval is set to -1
            and errorstr to appropriate error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        path = lmi_instance_to_path(instance)
        rval, rparams = self._cliconn.InvokeMethod(method, path, **params)
        return LMIReturnValue(rval=rval, rparams=rparams)

    @lmi_process_cim_exceptions_rval([])
    def get_associator_names(self, instance, AssocClass=None, ResultClass=None,
                             Role=None, ResultRole=None, limit=-1):
        """
        Returns a list of associated :py:class:`wbem.CIMInstanceName` objects
        with an input instance.

        :param instance: for this object the list of associated
            :py:class:`wbem.CIMInstanceName` will be returned. The object needs
            to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :param int limit: unused
        :returns: list of associated :py:class:`wbem.CIMInstanceName` objects
            with an input instance, if no error occurs; otherwise an empty list
            is returned
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        return self._cliconn.AssociatorNames(
            lmi_instance_to_path(instance),
            **filter_kwargs(
                AssocClass=AssocClass,
                ResultClass=ResultClass,
                Role=Role,
                ResultRole=ResultRole))

    @lmi_process_cim_exceptions_rval([])
    def get_associators(self, instance, AssocClass=None, ResultClass=None,
                        Role=None, ResultRole=None, IncludeQualifiers=False,
                        IncludeClassOrigin=False, PropertyList=None,
                        limit=-1):
        """
        Returns a list of associated :py:class:`wbem.CIMInstance` objects with
        an input instance.

        :param instance: for this object the list of associated
            :py:class:`wbem.CIMInstance` objects will be returned. The object
            needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of objects by mandating that each
            returned object shall be associated to the source object through an
            instance of this class or one of its subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be either an instance of this class (or one of its
            subclasses) or be this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall be associated with the source object through an association
            in which the source object plays the specified role. That is, the
            name of the property in the association class that refers to the
            source object shall match the value of this parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be associated to the source object through an association in
            which the returned object plays the specified role. That is, the
            name of the property in the association class that refers to the
            returned object shall match the value of this parameter.
        :param bool IncludeQualifiers: indicates, if all qualifiers for each
            object (including qualifiers on the object and on any returned
            properties) shall be included as ``<QUALIFIER>`` elements in the
            response.
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each
            returned object.
        :param list PropertyList: if not None, the members of the array define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If it is None, no additional filtering is defined.
        :param int limit: unused
        :returns: list of associated :py:class:`wbem.CIMInstance` objects with
            an input instance, if no error occurs; otherwise an empty list is
            returned
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        return self._cliconn.Associators(
            lmi_instance_to_path(instance),
            **filter_kwargs(
                AssocClass=AssocClass,
                ResultClass=ResultClass,
                Role=Role,
                ResultRole=ResultRole,
                IncludeQualifiers=IncludeQualifiers,
                IncludeClassOrigin=IncludeClassOrigin,
                PropertyList=PropertyList))

    @lmi_process_cim_exceptions_rval([])
    def get_reference_names(self, instance, ResultClass=None, Role=None,
                            limit=-1):
        """
        Returns a list of association :py:class:`wbem.CIMInstanceName` objects
        with an input instance.

        :param instance: for this object the association
            :py:class:`wbem.CIMInstanceName` objects will be returned. The
            object needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of object names by mandating that each returned
            Object Name identify an instance of this class (or one of its
            subclasses) or this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of object names by mandating that each returned object
            name shall identify an object that refers to the target instance
            through a property with a name that matches the value of this
            parameter.
        :param int limit: unused
        :returns: list of association :py:class:`wbem.CIMInstanceName` objects
            with an input instance, if no error occurs; otherwise an empty list
            is returned
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        return self._cliconn.ReferenceNames(
            lmi_instance_to_path(instance),
            **filter_kwargs(
                ResultClass=ResultClass,
                Role=Role))

    @lmi_process_cim_exceptions_rval([])
    def get_references(self, instance, ResultClass=None, Role=None,
                       IncludeQualifiers=False, IncludeClassOrigin=False,
                       PropertyList=None, limit=-1):
        """
        Returns a list of association :py:class:`wbem.CIMInstance` objects with
        an input instance.

        :param instance: for this object the list of association
            :py:class:`wbem.CIMInstance` objects will be returned. The object
            needs to be instance of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be an instance of this class (or one of its subclasses) or
            this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall refer to the target object through a property with a name
            that matches the value of this parameter.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers
            for each object (including qualifiers on the object and on any
            returned properties) shall be included as ``<QUALIFIER>``
            elements in the response.
        :param bool IncludeClassOrigin: bool flag indicating, if the
            ``CLASSORIGIN`` attribute shall be present on all appropriate
            elements in each returned object.
        :param list PropertyList: if not None, the members of the list define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If PropertyList is None, no additional filtering is
            defined.
        :param int limit: unused
        :returns: list of association :py:class:`wbem.CIMInstance` objects with
            an input instance, if no error occurs; otherwise an empty list is
            returned
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        return self._cliconn.References(
            lmi_instance_to_path(instance),
            **filter_kwargs(
                ResultClass=ResultClass,
                Role=Role,
                IncludeQualifiers=IncludeQualifiers,
                IncludeClassOrigin=IncludeClassOrigin,
                PropertyList=PropertyList))

    @lmi_process_cim_exceptions(None)
    def create_instance(self, classname, namespace=None, host=None,
                        properties=None, qualifiers=None, property_list=None):
        """
        Creates a new :py:class:`wbem.CIMInstance` object.

        :param string classname: class name of a new instance
        :param string namespace: namespace, of the new instance
        :param dictionary properties: property names and values
        :param dictionary qualifiers: qualifier names and values
        :param list property_list: list for property filtering; see
            :py:class:`wbem.CIMInstance`
        :returns: new :class:`wbem.CIMInstance`, if no error occurs; otherwise
            None is returned
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        # Create a new dictionaries from the input ones, we do not want to
        # modify user's input variables.
        properties = dict(properties) if properties is not None else {}
        qualifiers = dict(qualifiers) if qualifiers is not None else {}
        cim_instance = wbem.CIMInstance(
            classname,
            properties,
            qualifiers,
            wbem.CIMInstanceName(
                classname,
                namespace=namespace,
                host=host),
            property_list)
        cim_path = self._cliconn.CreateInstance(NewInstance=cim_instance)
        return self.get_instance(cim_path, LocalOnly=False)

    @lmi_process_cim_exceptions(-1)
    def modify_instance(self, instance, IncludeQualifiers=True,
                        PropertyList=None):
        """
        Modifies a :py:class:`wbem.CIMInstance` object at CIMOM side.

        :param wbem.CIMInstance instance: object to be modified
        :param bool IncludeQualifiers: indicates, if the qualifiers are
            modified as specified in `ModifiedInstance`.
        :param list PropertyList: if not None, the members of the list define
            one or more property names. Only properties specified in the
            PropertyList are modified.  Properties of the *ModifiedInstance*
            that are missing from the PropertyList are ignored. If the
            PropertyList is an empty list, no properties are modified.  If the
            PropertyList is None, the set of properties to be modified consists
            of those of *ModifiedInstance* with values different from the
            current values in the instance to be modified.
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to 0,
            if no error occurs; otherwise ``rval`` is set to -1 and
            ``errorstr`` is set to corresponding error string.
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        self._cliconn.ModifyInstance(
            instance,
            **filter_kwargs(
                IncludeQualifiers=IncludeQualifiers,
                PropertyList=PropertyList))
        return LMIReturnValue(rval=0)

    @lmi_process_cim_exceptions(False)
    def delete_instance(self, instance):
        """
        Deletes a :py:class:`wbem.CIMInstance` from the CIMOM side.

        :param instance: object to be deleted. The object needs to be instance
            of following classes:

            * :py:class:`wbem.CIMInstance`
            * :py:class:`wbem.CIMInstanceName`
            * :py:class:`.LMIInstance`
            * :py:class:`.LMIInstanceName`

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to True,
            if no error occurs; otherwise ``rval`` is set to False and
            ``errorstr`` is set to corresponding error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`,
            :py:exc:`TypeError`
        """
        self._cliconn.DeleteInstance(lmi_instance_to_path(instance))
        return LMIReturnValue(rval=True)

    @lmi_process_cim_exceptions()
    def exec_query(self, query_lang, query, namespace=None):
        """
        Executes a query and returns a list of :py:class:`wbem.CIMInstance`
        objects.

        :param string query_lang: query language
        :param string query: query to execute
        :param string namespace: target namespace for the query
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to list
            of :py:class:`wbem.CIMInstance` objects, if no error occurs;
            otherwise ``rval`` is set to None and ``errorstr`` is set to
            corresponding error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        inst_list = self._cliconn.ExecQuery(query_lang, query, namespace)
        return LMIReturnValue(rval=inst_list)

    @property
    def username(self):
        """
        :returns: user name as a part of provided credentials
        :rtype: string
        """
        return self._cliconn.creds[0]

    @property
    def uri(self):
        """
        :returns: URI of the CIMOM
        :rtype: string
        """
        return self._uri

    @property
    def hostname(self):
        """
        :returns: hostname of CIMOM
        :rtype: string
        """
        return self._cliconn.hostname



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Defines a base class for all command classes operating upon a
:py:class:`~lmi.scripts.common.session.Session` object.
"""
import abc
from collections import OrderedDict

from lmi.scripts.common import get_logger
from lmi.scripts.common.command import meta
from lmi.scripts.common.command.endpoint import LmiEndPointCommand
from lmi.scripts.common.configuration import Configuration
from lmi.scripts.common.session import Session
from lmi.shell import LMIConnection
from lmi.shell import LMIUtil

LOG = get_logger(__name__)

class LmiSessionCommand(LmiEndPointCommand):
    """
    Base class for end-point commands operating upon a session object.

    Using metaclass: :py:class:`~.meta.SessionCommandMetaClass`.
    """
    __metaclass__ = meta.SessionCommandMetaClass

    @classmethod
    def cim_namespace(cls):
        """
        This returns default cim namespace, the connection object will be
        nested into before being passed to associated function.
        It can be overriden in few ways:

            1. by setting ``[CIM] Namespace`` option in configuration
            2. by giving ``--namespace`` argument on command line to the
               ``lmi`` meta-command
            3. by setting ``NAMESPACE`` property in declaration of command

        Higher number means higher priority.
        """
        return Configuration.get_instance().namespace

    @classmethod
    def dest_pos_args_count(cls):
        """
        There is a namespace/connection object passed as the first positional
        argument.
        """
        return LmiEndPointCommand.dest_pos_args_count.im_func(cls) + 1

    def process_session(self, session, args, kwargs):
        """
        Process each host of given session, call the associated command
        function, collect results and print it to standard output.

        :param session: Session object with set of hosts.
        :type session: :py:class:`~lmi.scripts.common.session.Session`
        :param list args: Positional arguments to pass to associated function
            in command library.
        :param dictionary kwargs: Keyword arguments as a dictionary.
        :returns: Exit code of application.
        :rtype: integer
        """
        if not isinstance(session, Session):
            raise TypeError("session must be an object of Session, not %s"
                    % repr(session))
        # { ( hostname : (passed, result) ), ... }
        # where passed is a boolean and result is returned value if passed is
        # True and exception otherwise
        results = OrderedDict()
        for connection in session:
            try:
                result = self.take_action(connection, args, kwargs)
                # result may be a generator which may throw in the following
                # function
                results[connection.uri] = (True, result)
                self.process_host_result(connection.uri, True, result)
            except Exception as exc:
                if len(session) > 1:
                    LOG().exception('Invocation failed for host "%s": %s',
                            connection.uri, exc)
                else:
                    LOG().exception(exc)
                results[connection.uri] = (False, exc)
                self.process_host_result(connection.uri, False, exc)
        self.process_session_results(session, results)
        return all(r[0] for r in results.values())

    def process_host_result(self, hostname, success, result):
        """
        Called from :py:meth:`process_session` after single host gets
        processed. By default this prints obtained *result* with default
        formatter if the execution was successful. Children of this class may
        want to override this.

        :param string hostname: Name of host involved.
        :param boolean success: Whether the action on host succeeded.
        :param result: Either the value returned by associated function upon a
            successful invocation or an exception.
        """
        if success:
            if len(self.session) > 1:
                self.formatter.print_host(hostname)
            self.produce_output(result)

    def process_session_results(self, session, results):
        """
        Called at the end of :py:meth:`process_session`'s execution. It's
        supposed to do any summary work upon results from all hosts. By default
        it just prints errors in a form of list.

        :param session: Session object.
        :type session: :py:class:`lmi.scripts.common.session.Session`
        :param dictionary results: Dictionary of form: ::

                { 'hostname' : (success_flag, result), ... }

            where *result* is either an exception or returned value of
            associated function, depending on *success_flag*. See the
            :py:meth:`process_host_result`.
        """
        if not isinstance(session, Session):
            raise TypeError("session must be a Session object")
        if not isinstance(results, dict):
            raise TypeError("results must be a dictionary")
            # check whether any output has been produced
        if (  len(session.get_unconnected())
           or any(not r[0] for r in results.values())):
            data = []
            for hostname in session.get_unconnected():
                data.append((hostname, ['failed to connect']))
            for hostname, (success, error) in results.items():
                if not success:
                    data.append((hostname, [error]))
            if len(session) > 1:
                self._print_errors(data,
                        new_line=any(r[0] for r in results.values()))

    @abc.abstractmethod
    def take_action(self, connection, args, kwargs):
        """
        Executes an action on single host and collects results.

        :param connection: Connection to a single host.
        :type connection: :py:class:`lmi.shell.LMIConnection`
        :param list args: Positional arguments for associated function.
        :param dictionary kwargs: Keyword arguments for associated function.
        :returns: Column names and item list as a pair.
        :rtype: tuple
        """
        raise NotImplementedError("take_action must be implemented in subclass")

    def execute_on_connection(self, connection, *args, **kwargs):
        """
        Wraps the :py:meth:`~.endpoint.LmiEndPointCommand.execute` method with
        connection adjustments. Connection object is not usually passed
        directly to associated function. Mostly it's the namespace object that
        is expected. This method checks, whether the namespace object is
        desired and modifies connection accordingly.

        :param connection: Connection to single host.
        :type connection: :py:class:`lmi.shell.LMIConnection`
        :param list args: Arguments handed over to associated function.
        :param dictionary kwargs: Keyword arguments handed over to associated
            function.
        """
        if not isinstance(connection, LMIConnection):
            raise TypeError("expected an instance of LMIConnection for"
                    " connection argument, not %s" % repr(connection))
        namespace = self.cim_namespace()
        if namespace is not None:
            connection = LMIUtil.lmi_wrap_cim_namespace(
                    connection, namespace)
        return self.execute(connection, *args, **kwargs)

    def run_with_args(self, args, kwargs):
        return self.process_session(self.session, args, kwargs)




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module containing help command.
"""

from lmi.scripts.common import get_logger
from lmi.scripts.common.command import LmiEndPointCommand
from lmi.scripts.common import errors

LOG = get_logger(__name__)

EXIT_CODE_SUCCESS = 0
EXIT_CODE_FAILURE = 1
EXIT_CODE_KEYBOARD_INTERRUPT = 2
EXIT_CODE_COMMAND_NOT_FOUND = 3
EXIT_CODE_INVALID_SYNTAX = 4
EXIT_CODE_UNSATISFIED_DEPENDENCIES = 5

def _execute_exit(exit_code):
    """ Associated function with ``Exit`` command. """
    raise errors.LmiTerminate(exit_code)

class Exit(LmiEndPointCommand):
    """
    Terminate the shell.

    Usage: %(cmd)s [<exit_code>]
    """
    CALLABLE = _execute_exit
    OWN_USAGE = True

    def verify_options(self, options):
        code = options['<exit_code>']
        if code is not None:
            try:
                int(code)
            except ValueError:
                raise errors.LmiInvalidOptions(
                        "<exit_code> must be an integer not \"%s\"" % code)

    def transform_options(self, options):
        code = options.get('<exit_code>', None)
        if code is None:
            code = EXIT_CODE_SUCCESS
        options['<exit_code>'] = int(code)




# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import logging
import sys

from lmi.shell.LMIShellOptions import LMIShellOptions


class LMIShellLogger(logging.Logger):
    """
    LMIShell's logger with queueing capability.
    """
    def __init__(self, name, level=logging.NOTSET):
        # Due to Python2.6 compatibility, we call superclass c-tor in this way.
        logging.Logger.__init__(self, name, level)
        self.queue = []

    def _log(self, level, msg, args, exc_info=None, extra=None):
        if logging._srcfile:
            try:
                fn, lno, func = self.findCaller()
            except ValueError:
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

        record = self.makeRecord(
            self.name, level, fn, lno, msg, args, exc_info, func, extra)
        if self.isEnabledFor(level):
            self.handle(record)
        else:
            self.queue.append(record)

    def setLevel(self, level):
        """
        Sets a logging level of this handler. If there are any log records
        stored in internal queue, they are also handled.

        :param int level: logging level
        """
        logging.Logger.setLevel(self, level)
        self.processQueue()

    def processQueue(self):
        """
        Logs all enabled log records stored in internal queue.
        """
        # Handle all enabled log records.
        for r in self.queue:
            if not self.isEnabledFor(r.levelno):
                continue
            self.handle(r)

        # Remove all handled records.
        self.queue = filter(
            lambda r: not self.isEnabledFor(r.levelno),
            self.queue)

    def debug(self, msg, *args, **kwargs):
        """
        Log a message with severity 'DEBUG'.
        """
        self._log(logging.DEBUG, msg, args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log a message with severity 'CRITICAL'.
        """
        self._log(logging.CRITICAL, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log a message with severity 'ERROR'.
        """
        self._log(logging.ERROR, msg, args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        """
        Log a message with severity 'ERROR' also with exception information.
        """
        kwargs["exc_info"] = 1
        self.error(msg, args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log a message with severity 'INFO'.
        """
        self._log(logging.INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log a message with severity 'WARNING'.
        """
        self._log(logging.WARNING, msg, args, **kwargs)

logging.setLoggerClass(LMIShellLogger)


def lmi_init_logger():
    """
    Initializes LMIShell's logging.
    """
    logging.basicConfig(format='%(levelname)s: %(message)s')


def lmi_get_logger():
    """
    Returns LMIShell's logger.

    :returns: logger
    """
    return logging.getLogger("lmishell")


def lmi_setup_logger(log_options):
    """
    Sets logging level.

    :param int log_options: level defined in :py:class:`.LMIShellOptions`
    """
    logger = lmi_get_logger()
    if log_options == LMIShellOptions._LOG_MORE_VERBOSE:
        # Print out all the log messages to stderr stream
        logger.setLevel(logging.DEBUG)
    elif log_options == LMIShellOptions._LOG_VERBOSE:
        # Print out only a set of log messages to stderr stream
        logger.setLevel(logging.INFO)
    elif log_options == LMIShellOptions._LOG_QUIET:
        # Quiet flag seen, drop all the log messages
        handlers = logger.handlers
        for handler in handlers:
            logger.removeHandler(handler)
        logger.addHandler(logging.NullHandler())
    else:
        # By default, print error messages to stderr stream
        logger.setLevel(logging.ERROR)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

import LMIExceptions

from lmi.shell.LMIBaseObject import LMIWrapperBaseObject

from lmi.shell.LMIUtil import lmi_wrap_cim_namespace
from lmi.shell.LMIUtil import lmi_wrap_cim_class
from lmi.shell.LMIUtil import lmi_transform_to_lmi


class LMINamespace(LMIWrapperBaseObject):
    """
    LMI class representing CIM namespace.

    :param LMIConnection conn: connection object
    :param string name: namespace name
    """
    def __init__(self, conn, name):
        super(LMINamespace, self).__init__(conn)
        self._name = name

    def __getattr__(self, name):
        """
        Returns a :py:class:`.LMIClass` object.

        :param string name: class name
        :returns: :py:class:`.LMIClass` object
        """
        return lmi_wrap_cim_class(self._conn, name, self.name)

    def __repr__(self):
        """
        :returns: pretty string for the object
        """
        return "%s(namespace='%s', ...)" % (self.__class__.__name__, self.name)

    def classes(self):
        """
        Returns a list of class names.

        :param string filter_key: substring of a class name
        :param bool exact_match: tells, if to search for exact match or
            substring
        :returns: list of class names

        **Usage:** :ref:`namespaces_available_classes`.
        """
        class_names, _, errorstr = self._conn.client.get_class_names(
            self._name, DeepInheritance=True)
        if not class_names:
            return []
        return class_names

    def get_class(self, classname):
        """
        Returns :py:class:`.LMIClass`.

        :param string classname: class name of new :py:class:`.LMIClass`
        :raises: :py:exc:`.LMIClassNotFound`
        """
        class_names, _, _ = self._conn.client.get_class_names(
            self._name, DeepInheritance=True)
        if class_names is not None and classname in class_names:
            return lmi_wrap_cim_class(self._conn, classname, self.name)
        raise LMIExceptions.LMIClassNotFound(self.name, classname)

    def print_classes(self):
        """
        Prints out a list of classes.

        :param string filter_key: substring of a class name
        :param bool exact_match: tells, if to search for exact match, or to
            search for a matching substring

        **Usage:** :ref:`namespaces_available_classes`.
        """
        for c in self.classes():
            sys.stdout.write("%s\n" % c)

    def cql(self, query):
        """
        Executes a CQL query and returns a list of :py:class:`.LMIInstance`
        objects.

        :param string query: CQL query to execute
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of :py:class:`.LMIInstance` objects

        **Usage:** :ref:`namespaces_queries`.
        """
        inst_list, _, errorstr = self._conn.client.exec_query(
            self._conn.client.QUERY_LANG_CQL, query, self._name)
        if not inst_list:
            return []
        return lmi_transform_to_lmi(self._conn, inst_list)

    def wql(self, query):
        """
        Executes a WQL query and returns a list of :py:class:`.LMIInstance`
        objects.

        :param string query: WQL query to execute
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of :py:class:`.LMIInstance` objects

        **Usage:** :ref:`namespaces_queries`.
        """
        inst_list, _, errorstr = self._conn.client.exec_query(
            self._conn.client.QUERY_LANG_WQL, query, self._name)
        if not inst_list:
            return []
        return lmi_transform_to_lmi(self._conn, inst_list)

    @property
    def name(self):
        """
        :returns: namespace name
        :rtype: string
        """
        return self._name


class LMINamespaceRoot(LMINamespace):
    """
    Derived class for *root* namespace. Object of this class is accessible from
    :py:class:`.LMIConnection` object as a hierarchy entry.

    :param LMIConnection conn: connection object
    """
    _SUPPORTED_NAMESPACES = [
        "cimv2",
        "dcim",
        "interop",
        "PG_InterOp",
        "PG_Internal"
    ]

    def __init__(self, conn):
        super(LMINamespaceRoot, self).__init__(conn, "root")

    def __getattr__(self, name):
        if name in LMINamespaceRoot._SUPPORTED_NAMESPACES:
            return lmi_wrap_cim_namespace(self._conn, "root/" + name)
        raise AttributeError(name)

    @property
    def namespaces(self):
        """
        :returns: list of strings with available namespaces

        **Usage:** :ref:`namespaces_available_namespaces`.
        """
        return LMINamespaceRoot._SUPPORTED_NAMESPACES

    def print_namespaces(self):
        """
        Prints out all available namespaces accessible via the namespace
        `root`.

        **Usage:** :ref:`namespaces_available_namespaces`.
        """
        for ns in LMINamespaceRoot._SUPPORTED_NAMESPACES:
            sys.stdout.write("%s\n" % ns)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from lmi.shell.LMICIMXMLClient import LMICIMXMLClient
from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMIShellCache import LMIShellCache


class LMIShellClient(LMICIMXMLClient):
    """
    :class:`.LMIShellClient` overrides few methods due to caching purposes.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be
        performed
    :param string password: user's password
    :param bool interactive: flag indicating, if the LMIShell client is running
        in the interactive mode; default value is False.
    :param bool use_cache: flag indicating, if the LMIShell client should use
        cache for :class:`CIMClass` objects. This saves a lot's of
        communication, if there is often the
        :meth:`.LMIShellClient.get_class_names()` or
        :meth:`.LMIShellClient.attr.get_class()` call issued.

    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: indicates, whether a server side
        certificate needs to be verified, if SSL used; default value is True

    **NOTE:** If interactive is set to True, LMIShell will:

    * prompt for username and password, if missing and connection via
      Unix socket can not be established.
    * use pager for the output of: :py:meth:`.LMIInstance.doc`,
      :py:meth:`.LMIClass.doc`, :py:meth:`.LMIInstance.tomof` and
      :py:meth:`.LMIMethod.tomof`
    """
    def __init__(self, uri, username="", password="", interactive=False,
                 use_cache=True, key_file=None, cert_file=None,
                 verify_server_cert=True):
        super(LMIShellClient, self).__init__(
            uri, username, password, key_file=key_file,
            cert_file=cert_file, verify_server_cert=verify_server_cert)
        self._interactive = interactive
        self._cache = LMIShellCache(use_cache)

    def get_class_names(self, namespace=None, ClassName=None,
                        DeepInheritance=False):
        """
        Returns a list of class names.

        :param string namespace: namespace, from which the class names list
            should be retrieved; if None, default namespace will be used
            (**NOTE:** see :py:mod:`wbem`)
        :param string ClassName: defines the class that is the basis for the
            enumeration.  If the ClassName input parameter is absent, this
            implies that the names of all classes.
        :param bool DeepInheritance: if not present, of False, only the names
            of immediate child subclasses are returned, otherwise the names of
            all subclasses of the specified class should be returned.
        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to a
            list of strings containing class names, if no error occurs;
            otherwise ``rval`` is set to None and ``errorstr`` contains an
            appropriate error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        if self._cache.active:
            class_list = self._cache.get_classes(namespace)
            if class_list is None:
                class_list, _, errorstr = LMICIMXMLClient.get_class_names(
                    self, namespace, ClassName, DeepInheritance)
                if not class_list:
                    return LMIReturnValue(rval=class_list, errorstr=errorstr)
                self._cache.set_classes(class_list, namespace)
            return LMIReturnValue(rval=class_list)
        return LMICIMXMLClient.get_class_names(
            self, namespace, ClassName, DeepInheritance)

    def get_class(self, classname, namespace=None, LocalOnly=True,
                  IncludeQualifiers=True, IncludeClassOrigin=False,
                  PropertyList=None, full_fetch=False):
        """
        Returns a :py:class:`wbem.CIMClass` object.

        :param string classname: class name
        :param string namespace: -- namespace name, from which the
            :py:class:`wbem.CIMClass` should be retrieved; if None, default
            namespace will be used (**NOTE:** see :py:mod:`wbem`)
        :param bool LocalOnly: indicates, if only local members should be
            present in the returned :py:class:`wbem.CIMClass`; any CIM elements
            (properties, methods, and qualifiers), except those added or
            overridden in the class as specified in the classname input
            parameter, shall not be included in the returned class.
        :param bool IncludeQualifiers: indicates, if qualifiers for the class
            (including qualifiers on the class and on any returned properties,
            methods, or method parameters) shall be included in the response.
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in the
            returned class.
        :param list PropertyList: if present and not None, the members of the
            list define one or more property names. The returned class shall
            not include elements for properties missing from this list. Note
            that if LocalOnly is specified as True, it acts as an additional
            filter on the set of properties returned. For example, if property
            A is included in the PropertyList but LocalOnly is set to True and
            A is not local to the requested class, it is not included in the
            response. If the PropertyList input parameter is an empty list, no
            properties are included in the response. If the PropertyList input
            parameter is None, no additional filtering is defined.
        :returns: :py:class:`.LMIReturnValue` object with rval set to
            :py:class:`wbem.CIMClass`, if no error occurs; otherwise rval is
            set to none and errorstr to appropriate error string
        :raises: :py:exc:`.CIMError`, :py:exc:`.ConnectionError`
        """
        if full_fetch:
            IncludeQualifiers = full_fetch
            IncludeClassOrigin = full_fetch

        if self._cache.active:
            cls_cache_entry = self._cache.get_class(classname, namespace)
            if cls_cache_entry is None or \
                    (not cls_cache_entry.full_fetch and full_fetch):
                cls, _, errorstr = LMICIMXMLClient.get_class(
                    self, classname, namespace, LocalOnly, IncludeQualifiers,
                    IncludeClassOrigin, PropertyList)
                if cls is None:
                    return LMIReturnValue(rval=cls, errorstr=errorstr)
                self._cache.add_class(cls, namespace, full_fetch)
                self._cache.add_superclass(
                    cls.classname, cls.superclass, namespace)
            else:
                cls = cls_cache_entry.cim_class
            return LMIReturnValue(rval=cls)
        return LMICIMXMLClient.get_class(
            self, classname, namespace, LocalOnly, IncludeQualifiers,
            IncludeClassOrigin, PropertyList)

    def get_superclass(self, classname, namespace=None):
        """
        Returns string of a superclass to given class, if such superclass
        exists, None otherwise.

        :param string classname: class name
        :param string namespace: namespace name
        :returns: superclass name to a given classname or None
        :raises: :exc:`CIMError`, :exc:`ConnectionError`
        """
        if self._cache.active:
            rparams = None
            errorstr = ""
            if self._cache.has_superclass(classname, namespace):
                superclass = self._cache.get_superclass(classname, namespace)
            else:
                superclass, rparams, errorstr = LMICIMXMLClient.get_superclass(
                    self, classname, namespace)
                self._cache.add_superclass(classname, superclass, namespace)
            return LMIReturnValue(
                rval=superclass, rparams=rparams, errorstr=errorstr)
        return LMICIMXMLClient.get_superclass(self, classname, namespace)

    @property
    def interactive(self):
        """
        :returns: flag, if the LMIShell is run in the interactive mode
        :rtype: bool
        """
        return self._interactive

    @interactive.setter
    def interactive(self, i):
        """
        Property setter for interactive flag.

        :param bool i: new :attr:`.LMIShellClient.interactive` state
        """
        self._interactive = bool(i)

    @property
    def cache(self):
        """
        :returns: LMIShell's cache
        :rtype: :py:class:`.LMIShellCache`
        """
        return self._cache

    @property
    def use_cache(self):
        """
        :returns: flag, which tells, if the LMIShell should use a cache
        :rtype: bool
        """
        return self._cache.active

    @use_cache.setter
    def use_cache(self, active):
        """
        Property setter for cache activity.

        :param bool active: new :attr:LMIShellClient.use_cache` state
        """
        self._cache.active = active



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module defining the root command (``lmi`` binary).
"""

USAGE_STRING = \
"""
OpenLMI command line interface for CIM providers. It's functionality is
composed of registered subcommands, operating on top of simple libraries,
interfacing with particular OpenLMI profile providers.
Works also in interactive mode which is entered, when <command> argument is
omitted.

Usage:
    %(cmd)s [options] [(--trace | --notrace)] [-v]... [-h <host>]...
        <command> [<args> ...]
    %(cmd)s [options] [(--trace | --notrace)] [-v]... [-h <host>]...
    %(cmd)s (--help | --version)

Options:
    -c --config-file <config>  Path to a user configuration file. Options
                               specified here override any settings of global
                               configuration file.
    -h --host <host>           Hostname of target system.
    --hosts-file <hosts>       Path to a file containing target hostnames.
                               Each hostname must be listed on a single line.
    --user <user>              Username used in connection to any target host.
    --same-credentials         Use the first credentials given for all hosts.
    -n --noverify              Do not verify cimom's ssl certificate.
    -v                         Increase verbosity of output.
    --trace                    Show tracebacks on errors.
    --notrace                  Suppress tracebacks for exceptions.
    -q --quiet                 Supress output except for errors.
    --log-file <log_file>      Output file for logging messages.
    --namespace <namespace>    Default CIM namespace to use.
    -N --no-headings           Don't print table headings.
    -H --human-friendly        Print large values in human friendly units (i.e.
                               MB, GB, TB etc.)
    -L --lister-format (table | csv)
                               Print output of lister commands in CSV or table
                               format. CSV format is more suitable for machine
                               processing. Defaults to table.
    --help                     Show this text and quit.
    --version                  Print version of '%(cmd)s' in use and quit.

Handling hosts:
    If no --host or --hosts-file are given, the "localhost" will be tried. When
    running under root with Pegasus CIMOM, this results in a connection over
    unix socket (no need to supply credentials).

    Hosts may contain embedded credentials e.g.:
        http://user:passwd@hostname:5988
    Avoid supplying them on command line though. Always prefer the --hosts-file
    option.
"""

import docopt

from lmi.scripts._metacommand import exit
from lmi.scripts._metacommand import util
from lmi.scripts._metacommand import Interactive
from lmi.scripts.common import get_logger
from lmi.scripts.common import errors
from lmi.scripts.common.command import base

LOG = get_logger(__name__)

class TopLevelCommand(base.LmiBaseCommand):
    """
    Top level (instance, without any parent) command handling application
    parameters and passing work to registered subcommands.
    """

    @classmethod
    def has_own_usage(cls):
        return True

    @classmethod
    def child_commands(cls):
        return []

    @classmethod
    def is_end_point(cls):
        return False

    def __init__(self, app, cmd_name='lmi'):
        base.LmiBaseCommand.__init__(self, app, cmd_name)

    def get_usage(self, proper=False):
        return USAGE_STRING[1:] % { 'cmd' : " ".join(self.cmd_name_parts) }

    def run_subcommand(self, cmd_name, args):
        """
        Finds a command factory, instantiates it and passes the control.
        """
        cmd_factory = self.app.command_manager[cmd_name]
        cmd = cmd_factory(self.app, cmd_name, parent=self)
        return cmd.run(args)

    def start_interactive_mode(self):
        """ Run the command line loop of interactive application. """
        self.app.command_manager.add_command("exit", exit.Exit)
        iapp = Interactive(self)
        while True:
            try:
                ret = iapp.cmdloop()
                break
            except errors.LmiTerminate as err:
                ret = err.args[0]
                break
            except KeyboardInterrupt as err:
                LOG().debug('%s: %s', err.__class__.__name__, str(err))
                self.app.stdout.write('\n')
        iapp.save_history()
        return ret

    def run(self, args):
        """
        Handle program arguments, set up the application and call
        a subcommand or enter interactive mode. Return exit code.

        :param args: (``list``) Arguments without the binary name.
        """
        if not isinstance(args, (tuple, list)):
            raise TypeError("args must be a list")
        try:
            options = docopt.docopt(self.get_usage(), args,
                    version=util.get_version(), help=False, options_first=True)
        except docopt.DocoptLanguageError as exc:
            self.app.stderr.write("%s\n" % str(exc))
            return exit.EXIT_CODE_FAILURE
        if options.pop('--help', False):
            self.app.stdout.write(self.get_usage())
            self.app.stdout.write("\nCommands:\n")
            self.app.stdout.write("    %s\n" % " ".join(
                n for n in sorted(self.app.command_manager)))
            return exit.EXIT_CODE_SUCCESS
        if options.pop('--version', False):
            self.app.print_version()
            return exit.EXIT_CODE_SUCCESS
        self.app.setup(options)
        if options['<command>'] is None:
            return self.start_interactive_mode()

        try:
            LOG().debug('Running command "%s".', options['<command>'])
            return self.run_subcommand(options['<command>'], options['<args>'])
        except docopt.DocoptExit as err:
            if '--help' in args:
                cmd_args = options['<args>']
                cmd_args = cmd_args[:cmd_args.index('--help')]
                return self.run_subcommand(
                        'help', [options['<command>']] + cmd_args)
            raise





# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import abc
import sys
import re

from lmi.shell.LMIShellLogger import lmi_get_logger
from lmi.shell.LMIUtil import lmi_cast_to_lmi

logger = lmi_get_logger()


class LMIConstantValues(object):
    """
    Abstract class for constant value objects.

    :param cim_obj: this object is either of type
        :py:class:`wbem.CIMParameter`, :py:class:`wbem.CIMProperty` or
        :py:class:`wbem.CIMMethod`. Construction of this object requires to
        have a member ``_cast_type`` to properly cast CIM object. When
        constructing derived objects, make sure, that the mentioned member is
        present before calling this constructor.
    :param cast_type: parameter/property cast type
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, cim_obj, cast_type):
        items = zip(
            cim_obj.qualifiers["Values"].value,
            cim_obj.qualifiers["ValueMap"].value)
        self._value_map = {}
        self._value_map_inv = {}
        self._cast_type = cast_type
        # Fill two dictionaries for bidirectional access to constant values.
        cnt = 1
        for key, value in items:
            try:
                # Cast constant value first. If we get ValueError, no key
                # modifications are necessary.
                val = lmi_cast_to_lmi(self._cast_type, value)

                # Keys can contain various undesirable characters, such as
                # python operators, etc. So we drop them.
                mod_key = re.sub("\W", "", key)
                if mod_key[0].isdigit():
                    mod_key = "Key_" + mod_key
                if mod_key in self._value_map:
                    mod_key += str(cnt)
                    cnt += 1
                    logger.warn("Constant value mapped as: '%s' -> '%s'" %
                                (key, mod_key))

                self._value_map[mod_key] = val
                # For inverse mapping, we use unmodified key.
                self._value_map_inv[val] = key
            except ValueError, e:
                # Can not cast such value as interval. Can be found in
                # DMTFReserved, VendorReserved values.
                pass

    def __repr__(self):
        """
        Returns a string of all constant names with corresponding value.

        :returns: pretty string
        """
        result = ""
        for k, v in self._value_map.iteritems():
            result += "%s = %s\n" % (k, v)
        return result[:-1]

    def __getattr__(self, name):
        """
        Returns either a member of the class, or a constant value.

        Simplifies the code and constant value can be retrieved by
        :samp:`object.constant_value`.

        :param string name: member to retrieve
        :returns: class member
        """
        if name in self._value_map:
            return self._value_map[name]
        raise AttributeError(name)

    def print_values(self):
        """
        Prints all available constant names.

        **Usage:** :ref:`class_get_valuemap_properties`.
        """
        for k in self._value_map.keys():
            sys.stdout.write("%s\n" % k)

    def values_dict(self):
        """
        :returns: dictionary of constants' names and values
        """
        return self._value_map

    def values(self):
        """
        :returns: list of all available constant values
        """
        return self._value_map.keys()

    def value(self, value_name):
        """
        :param string value_name: constant name
        :returns: constant value

        **Usage:** :ref:`class_get_valuemap_property_value`.
        """
        return getattr(self, value_name)

    def value_name(self, value):
        """
        :param int value: numeric constant value
        :returns: constant value
        :rtype: string

        **Usage:** :ref:`class_get_valuemap_property_name`.
        """
        return self._value_map_inv[value]


class LMIConstantValuesParamProp(LMIConstantValues):
    """
    Derived class used for constant values of :py:class:`wbem.CIMProperty` and
    :py:class:`wbem.CIMParameter`.

    :param cim_property: :py:class:`wbem.CIMProperty` or
        :py:class:`wbem.CIMParameter` object. Both objects have necessary
        member ``type`` which is needed for proper casting.
    """
    def __init__(self, cim_property):
        super(LMIConstantValuesParamProp, self).__init__(
            cim_property, cim_property.type)


class LMIConstantValuesMethodReturnType(LMIConstantValues):
    """
    Derived class used for constant values of :py:class:`wbem.CIMMethod`.

    :param CIMMethod cim_method: :py:class:`wbem.CIMMethod` object
    """
    def __init__(self, cim_method):
        super(LMIConstantValuesMethodReturnType, self).__init__(
            cim_method, cim_method.return_type)



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Defines base command class for all endpoint commands. Those having no children.
"""
import abc
import inspect
import re
from docopt import docopt

from lmi.scripts.common import errors
from lmi.scripts.common import formatter
from lmi.scripts.common import get_logger
from lmi.scripts.common.formatter import command as fcmd
from lmi.scripts.common.command import base
from lmi.scripts.common.command import meta
from lmi.scripts.common.command import util

LOG = get_logger(__name__)

def opt_name_sanitize(opt_name):
    """
    Make a function parameter name out of option name. This replaces any
    character not suitable for python identificator with ``'_'`` and
    make the whole string lowercase.

    :param string opt_name: Option name.
    :returns: Modified option name.
    :rtype: string
    """
    return re.sub(r'[^a-zA-Z0-9]+', '_', opt_name).lower()

def options_dict2kwargs(options):
    """
    Convert option name from resulting ``docopt`` dictionary to a valid python
    identificator token used as function argument name.

    :param dictionary options: Dictionary returned by docopt call.
    :returns:  New dictionary with keys passable to function as argument
        names.
    :rtype: dictionary
    """
    # (new_name, value) for each pair in options dictionary
    kwargs = {}
    # (new_name, name)
    orig_names = {}
    for name, value in options.items():
        for (reg, func) in (
                (util.RE_OPT_BRACKET_ARGUMENT, lambda m: m.group('name')),
                (util.RE_OPT_UPPER_ARGUMENT,   lambda m: m.group('name')),
                (util.RE_OPT_SHORT_OPTION,     lambda m: m.group(0)),
                (util.RE_OPT_LONG_OPTION,      lambda m: m.group(0)),
                (util.RE_COMMAND_NAME,         lambda m: m.group(0))):
            match = reg.match(name)
            if match:
                new_name = func(match)
                break
        else:
            raise errors.LmiError(
                    'Failed to convert argument "%s" to function option.' %
                    name)
        if new_name == '--':
            continue    # ignore double dash
        new_name = opt_name_sanitize(new_name)
        if new_name in kwargs:
            raise errors.LmiError('Option clash for "%s" and "%s", which both'
                ' translate to "%s".' % (name, orig_names[new_name], new_name))
        kwargs[new_name] = value
        orig_names[new_name] = name
    return kwargs

class LmiEndPointCommand(base.LmiBaseCommand):
    """
    Base class for any leaf command.

    List of additional recognized properties:

        ``CALLABLE`` : ``tuple``
            Associated function. Will be wrapped in
            :py:meth:`LmiEndPointCommand.execute` method and will be accessible
            directly as a ``cmd.execute.dest`` property. It may be specified
            either as a string in form ``"<module_name>:<callable>"`` or as a
            reference to callable itself.
        ``ARG_ARRAY_SUFFIX`` : ``str``
            String appended to every option parsed by ``docopt`` having list as
            an associated value. It defaults to empty string. This modification
            is applied before calling
            :py:meth:`LmiEndPointCommand.verify_options` and
            :py:meth:`LmiEndPointCommand.transform_options`.
        ``FORMATTER`` : callable
            Default formatter factory for instances of given command. This
            factory accepts an output stream as the only parameter and returns
            an instance of :py:class:`~lmi.scripts.common.formatter.Formatter`.

    Using metaclass:
        :py:class:`.meta.EndPointCommandMetaClass`.
    """
    __metaclass__ = meta.EndPointCommandMetaClass

    def __init__(self, *args, **kwargs):
        super(LmiEndPointCommand, self).__init__(*args, **kwargs)
        self._formatter = None
        # saved options dictionary after call to transform_options()
        self._options = None

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """
        Subclasses must override this method to pass given arguments to
        command library function. This function shall be specified in
        ``CALLABLE`` property.
        """
        raise NotImplementedError("execute method must be overriden"
                " in subclass")

    def formatter_factory(self):
        """
        Subclasses shall override this method to provide default formatter
        factory for printing output.

        :returns: Subclass of basic formatter.
        """
        return formatter.Formatter

    @classmethod
    def dest_pos_args_count(cls):
        """
        Number of positional arguments the associated function takes from
        command. These arguments are created by the command alone -- they do
        not belong to options in usage string. Function can take additional
        positional arguments that need to be covered by usage string.

        :rtype: integer
        """
        dest = getattr(cls.execute, "dest", cls.execute)
        abstract = dest == cls.execute and util.is_abstract_method(
                cls, 'execute', True)
        # if the destination function is not yet defined (abstract is True)
        # let's assume it's not a method => 0 positional arguments needed
        return 1 if not abstract and inspect.ismethod(dest) else 0

    def run_with_args(self, args, kwargs):
        """
        Process end-point arguments and exit.

        :param list args: Positional arguments to pass to associated
            function in command library.
        :param dictionary kwargs: Keyword arguments as a dictionary.
        :returns: Exit code of application.
        :rtype: integer
        """
        return self.execute(*args, **kwargs)

    @property
    def formatter(self):
        """
        Return instance of default formatter.

        :rtype: :py:class:`~lmi.scripts.common.formatter.Formatter`
        """
        if self._formatter is None:
            opts = self.format_options
            factory = self.formatter_factory()
            argspec = inspect.getargspec(
                    factory.__init__ if type(factory) is type else factory)
            if not argspec.keywords:
                kwargs = dict((k, v) for k, v in opts.items() if k in argspec.args)
            self._formatter = factory(self.app.stdout, **kwargs)
        return self._formatter

    def _make_end_point_args(self, options):
        """
        Creates a pair of positional and keyword arguments for a call to
        associated function from command line options. All keyword
        options not expected by target function are removed.

        :param dictionary options: Output of ``docopt`` parser.
        :returns: Positional and keyword arguments as a pair.
        :rtype: tuple
        """
        # if execute method does not have a *dest* attribute, then it's
        # itself a destination
        dest = getattr(self.execute, "dest", self.execute)
        argspec = inspect.getargspec(dest)
        kwargs = options_dict2kwargs(options)
        # number of positional arguments not covered by usage string
        pos_args_count = self.dest_pos_args_count()
        to_remove = []
        # if associated function takes keyword arguments in a single
        # dictionary (kwargs), we can pass all options
        if argspec.keywords is None:
            # otherwise we need to remove any unhandled
            for opt_name in kwargs:
                if opt_name not in argspec.args[pos_args_count:]:
                    if opt_name not in self.cmd_name_parts:
                        LOG().debug('Option "%s" not handled in function "%s",'
                            ' ignoring.', opt_name, self.cmd_name)
                    to_remove.append(opt_name)
        for opt_name in to_remove:
            # remove options unhandled by function
            del kwargs[opt_name]
        args = []
        for arg_name in argspec.args[pos_args_count:]:
            if arg_name not in kwargs:
                raise errors.LmiCommandError(
                    self.__module__, self.__class__.__name__,
                    'registered command "%s" expects option "%s", which'
                    ' is not covered in usage string'
                    % (self.cmd_name, arg_name))
            args.append(kwargs.pop(arg_name))
        return args, kwargs

    def _preprocess_options(self, options):
        """
        This method may be overriden by
        :py:class:`~.meta.EndPointCommandMetaClass`
        as a result of processing ``ARG_ARRAY_SUFFIX`` and other properties
        modifying names of parsed options.

        This should not be overriden in command class's body.

        :param dictionary options: The result of ``docopt`` parser invocation
            which can be modified by this method.
        """
        pass

    def _parse_args(self, args):
        """
        Run ``docopt`` command line parser on given list of arguments.
        Removes all unrelated commands from created dictionary of options.

        :param list args: List of command line arguments just after the
            current command.
        :returns: Dictionary with parsed options. Please refer to
            docopt_ documentation for more informations.
        :rtype: dictionary

        .. _docopt: http://docopt.org/
        """
        full_args = self.get_cmd_name_parts(for_docopt=True) + args
        options = docopt(self.get_usage(), full_args, help=False)
        self._preprocess_options(options)

        # remove all command names from options
        cmd = self.parent
        while cmd is not None and not cmd.has_own_usage():
            cmd = cmd.parent
        if cmd is not None:
            for scn in cmd.child_commands():
                try:
                    del options[scn]
                except KeyError:
                    LOG().warn('Usage string of "%s.%s" command does not'
                            ' contain registered command "%s" command.',
                            cmd.__module__, cmd.__class__.__name__, scn)
        # remove also the root command name from options
        if cmd is not None and cmd.cmd_name in options:
            del options[cmd.cmd_name]
        return options

    def verify_options(self, options):
        """
        This method can be overriden in subclasses to check, whether the
        options given on command line are valid. If any flaw is discovered, an
        :py:exc:`~lmi.scripts.common.errors.LmiInvalidOptions` exception shall
        be raised. Any returned value is ignored.

        .. note::
            This is run before :py:meth:`transform_options()` method.

        :param dictionary options: Dictionary as returned by ``docopt`` parser.
        """
        pass

    def transform_options(self, options):
        """
        This method can be overriden in subclasses if options shall be somehow
        modified before passing them associated function.

        .. note::
            Run after :py:meth:`verify_options()` method.

        :param dictionary options: Dictionary as returned by ``docopt`` parser.
        """
        pass

    def produce_output(self, data):
        """
        This method can be use to render and print results with default
        formatter.

        :param data: Is an object expected by the
            :py:meth:`~lmi.scripts.common.formatter.Formatter.produce_output`
            method of formatter.
        """
        self.formatter.produce_output(data)

    def run(self, args):
        """
        Create options dictionary from input arguments, verify them,
        transform them, make positional and keyword arguments out of them and
        pass them to ``process_session()``.

        :param list args: List of command arguments.
        :returns: Exit code of application.
        :rtype: integer
        """
        options = self._parse_args(args)
        self.verify_options(options)
        self.transform_options(options)
        self._options = options.copy()
        args, kwargs = self._make_end_point_args(options)
        return self.run_with_args(args, kwargs)

    def _print_errors(self, error_list, new_line=True):
        """
        Print list of errors.

        :param list errors: Errors to print. Each error is a ``tuple``: ::

                (hostname, [error, error])

            Where ``error`` may be a test description or an instance of
            exception.
        :param new_line: Whether to print the new line before new error
            table is printed.
        """
        fmt = formatter.ErrorFormatter(self.app.stderr)
        if new_line:
            fmt.out.write('\n')
        if error_list:
            new_table_cmd = fcmd.NewTableCommand("There %s %d error%s" %
                ( 'were' if len(error_list) > 1 else 'was'
                , len(error_list)
                , 's' if len(error_list) > 1 else ''))
            fmt.produce_output((new_table_cmd, ))
        for hostname, host_errors in error_list:
            fmt.produce_output((fcmd.NewHostCommand(hostname), ))
            fmt.produce_output(host_errors)




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module containing help command.
"""

from lmi.scripts.common import errors
from lmi.scripts.common import get_logger
from lmi.scripts.common.command import LmiEndPointCommand
from lmi.scripts._metacommand import cmdutil
from lmi.scripts._metacommand import exit

LOG = get_logger(__name__)

class Help(LmiEndPointCommand):
    """
    Print the list of supported commands with short description.
    If a subcommand is given, its detailed help will be printed.

    Usage: %(cmd)s [<subcommand>...]
    """
    OWN_USAGE = True

    def execute(self, subcommand):
        mgr = self.app.command_manager
        node = self.app.active_command
        toplevel = self
        while toplevel.parent is not None:
            toplevel = toplevel.parent

        if node or subcommand:
            # Help for some subcommand will be printed.
            if node is None:
                node = toplevel
            if subcommand:
                index = 0
                try:
                    while index < len(subcommand) and not node.is_end_point():
                        while node.is_selector():
                            cmd_factory, _ = node.select_cmds().next()
                            node = cmd_factory(self.app, node.cmd_name,
                                    node.parent)
                        # selector may return either multiplexer or end-point
                        if node.is_end_point():
                            break
                        cmd_factory = cmdutil.get_subcommand_factory(node,
                                subcommand[index])
                        node = cmd_factory(self.app, subcommand[index], node)
                        index += 1
                except errors.LmiCommandNotFound as err:
                    LOG().error(str(err))
            if node is not toplevel:
                self.app.stdout.write(node.get_usage(True))
                if node is self.app.active_command:
                    # show additional information only when no command given
                    self.app.stdout.write('\nTo get help for built-in commands,'
                            ' type:\n    :help\n')
                return exit.EXIT_CODE_SUCCESS

        # let's print the summary of available commands
        self.app.stdout.write("Commands:\n")
        max_cmd_len = max(len(n) for n in mgr)
        cmd_line = "  %%-%ds - %%s\n" % max_cmd_len
        for cmd in sorted(mgr):
            self.app.stdout.write(cmd_line
                    % (cmd, mgr[cmd].get_description()
                        .strip().split("\n", 1)[0]))
        self.app.stdout.write(
                "\nFor more informations about particular command type:\n"
                "    help <command>\n")

        return exit.EXIT_CODE_SUCCESS




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Subpackage with formatter classes used to render and output results.

Each formatter has a :py:meth:`Formatter.produce_output` method taking one
argument which gets rendered and printed to output stream. Each formatter
expects different argument, please refer to doc string of particular class.
"""

import itertools
import locale
import os

from lmi.scripts.common import errors
from lmi.scripts.common.formatter import command as fcmd

def get_terminal_width():
    """
    Get the number of columns of current terminal if attached to it. It
    defaults to 79 characters.

    :returns: Number of columns of attached terminal.
    :rtype: integer
    """
    try:
        term_cols = int(os.popen('stty size', 'r').read().split()[1])
    except (IOError, OSError, ValueError):
        term_cols = 79  # fallback
    return term_cols

class Formatter(object):
    """
    Base formatter class.

    It produces string representation of given argument and prints it.

    This formatter supports following commands:

        * :py:class:`~.command.NewHostCommand`.

    :param file stream: Output stream.
    :param integer padding: Number of leading spaces to print at each line.
    :param boolean no_headings: If table headings should be omitted.
    """

    def __init__(self, stream, padding=0, no_headings=False):
        if not isinstance(padding, (int, long)):
            raise TypeError("padding must be an integer")
        if padding < 0:
            padding = 0
        self.out = stream
        self.padding = padding
        self.no_headings = no_headings
        #: counter of hosts printed
        self.host_counter = 0
        #: counter of tables produced for current host
        self.table_counter = 0
        #: counter of lines producted for current table
        self.line_counter = 0

    @property
    def encoding(self):
        """
        Try to determine encoding for output terminal.

        :returns: Encoding used to encode unicode strings.
        :rtype: string
        """
        enc = getattr(self.out, 'encoding')
        if not enc:
            enc = locale.getpreferredencoding()
        return enc

    def render_value(self, val):
        """
        Rendering function for single value.

        :param val: Any value to render.
        :returns: Value converted to its string representation.
        :rtype: str
        """
        if isinstance(val, unicode):
            return val.encode(self.encoding)
        if not isinstance(val, str):
            val = str(val)
        return val

    def print_line(self, line, *args, **kwargs):
        """
        Prints single line. Output message is prefixed with ``padding`` spaces,
        formated and printed to output stream.

        :param string line: Message to print, it can contain markers for
            other arguments to include according to ``format_spec`` language.
            Please refer to ``Format Specification Mini-Language`` in python
            documentation.
        :param list args: Positional arguments to ``format()`` function of
            ``line`` argument.
        :param dictionary kwargs: Keyword arguments to ``format()`` function.
        """
        line = ' ' * self.padding + line.format(*args, **kwargs)
        self.out.write(line.encode(self.encoding))
        self.out.write("\n")
        self.line_counter += 1

    def print_host(self, hostname):
        """
        Prints header for new host.

        :param string hostname: Hostname to print.
        """
        if (  self.host_counter > 0
           or self.table_counter > 0
           or self.line_counter > 0):
            self.out.write("\n")
        term_width = get_terminal_width()
        self.out.write("="*term_width + "\n")
        self.out.write("Host: %s\n" % hostname)
        self.out.write("="*term_width + "\n")
        self.host_counter += 1
        self.table_counter = 0
        self.line_counter = 0

    def produce_output(self, data):
        """
        Render and print given data.

        Data can be also instance of
        :py:class:`~.command.FormatterCommand`, see
        documentation of this class for list of
        allowed commands.

        This shall be overridden by subclasses.

        :param data: Any value to print. Subclasses may specify their
            requirements for this argument. It can be also am instance of
            :py:class:`~.command.FormatterCommand`.
        """
        self.print_line(str(data))
        self.line_counter += 1

class ListFormatter(Formatter):
    """
    Base formatter taking care of list of items. It renders input data in a
    form of table with mandatory column names at the beginning followed by
    items, one occupying single line (row).

    This formatter supports following commands:
        * :py:class:`~.command.NewHostCommand`
        * :py:class:`~.command.NewTableCommand`
        * :py:class:`~.command.NewTableHeaderCommand`

    The command must be provided as content of one row. This row is then not
    printed and the command is executed.

    This class should be subclassed to provide nice output.
    """
    def __init__(self, stream, padding=0, no_headings=False):
        super(ListFormatter, self).__init__(stream, padding, no_headings)
        self.column_names = None

    def print_text_row(self, row):
        """
        Print data row without any header.

        :param tuple row: Data to print.
        """
        self.out.write(" "*self.padding + self.render_value(row) + "\n")
        self.line_counter += 1

    def print_row(self, data):
        """
        Print data row. Optionaly print header, if requested.

        :param tuple data: Data to print.
        """
        if self.line_counter == 0 and not self.no_headings:
            self.print_header()
        self.print_text_row(data)

    def print_table_title(self, title):
        """
        Prints title of next tale.

        :param string title: Title to print.
        """
        if self.table_counter > 0 or self.line_counter > 0:
            self.out.write('\n')
        self.out.write("%s:\n" % title)
        self.table_counter += 1
        self.line_counter = 0

    def print_header(self):
        """ Print table header. """
        if self.no_headings:
            return
        if self.column_names:
            self.print_text_row(self.column_names)

    def produce_output(self, rows):
        """
        Prints list of rows.

        There can be a :py:class:`~.command.FormatterCommand` instance instead
        of a row. See documentation of this class for list of allowed commands.

        :param rows:  List of rows to print.
        :type rows: list, generator or :py:class:`.command.FormatterCommand`
        """
        for row in rows:
            if isinstance(row, fcmd.NewHostCommand):
                self.print_host(row.hostname)
            elif isinstance(row, fcmd.NewTableCommand):
                self.print_table_title(row.title)
            elif isinstance(row, fcmd.NewTableHeaderCommand):
                self.column_names = row.columns
            else:
                self.print_row(row)

class TableFormatter(ListFormatter):
    """
    Print nice human-readable table to terminal.

    To print the table nicely aligned, the whole table must be populated first.
    Therefore this formatter stores all rows locally and does not print
    them until the table is complete. Column sizes are computed afterwards
    and the table is printed at once.

    This formatter supports following commands:
        * :py:class:`~.command.NewHostCommand`
        * :py:class:`~.command.NewTableCommand`
        * :py:class:`~.command.NewTableHeaderCommand`

    The command must be provided as content of one row. This row is then not
    printed and the command is executed.
    """
    def __init__(self, stream, padding=0, no_headings=False):
        super(TableFormatter, self).__init__(stream, padding, no_headings)
        self.stash = []

    def print_text_row(self, row, column_sizes):
        for i in xrange(len(row)):
            size = column_sizes[i]
            # Convert to unicode to compute correct length of utf-8 strings
            # (e.g. with fancy trees with utf-8 graphics).
            item = (    unicode(row[i]) if not isinstance(row[i], str)
                   else row[i].decode(self.encoding))
            if i < len(row) - 1:
                item = item.ljust(size)
            self.out.write(self.render_value(item))
            if i < len(row) - 1:
                self.out.write(" ")
        self.out.write("\n")
        self.line_counter += 1

    def print_stash(self):
        # Compute column sizes
        column_sizes = []
        rows = iter(self.stash)
        try:
            if self.column_names is None:
                row = rows.next()
            else:
                row = self.column_names
            for i in xrange(len(row)):
                column_sizes.append(len(row[i]))
        except StopIteration:
            pass # empty stash

        for row in rows:
            for i in xrange(len(row)):
                row_length = len(
                         unicode(row[i]) if not isinstance(row[i], str)
                    else row[i].decode(self.encoding))
                if column_sizes[i] < row_length:
                    column_sizes[i] = row_length

        # print headers
        if not self.no_headings and self.column_names:
            self.print_text_row(self.column_names, column_sizes)
        # print stashed rows
        for row in self.stash:
            self.print_text_row(row, column_sizes)
        self.stash = []

    def print_row(self, data):
        """
        Print data row.

        :param tuple data: Data to print.
        """
        self.stash.append(data)

    def print_host(self, hostname):
        """
        Prints header for new host.

        :param string hostname: Hostname to print.
        """
        if len(self.stash):
            # without a check, this would print headers for empty stash
            self.print_stash()
        super(TableFormatter, self).print_host(hostname)

    def print_table_title(self, title):
        """
        Prints title of next tale.

        :param string title: Title to print.
        """
        if len(self.stash):
            # without a check, this would print headers for empty stash
            self.print_stash()
        if self.table_counter > 0 or self.line_counter > 0:
            self.out.write('\n')
        self.out.write("%s:\n" % title)
        self.table_counter += 1
        self.line_counter = 0

    def produce_output(self, rows):
        """
        Prints list of rows.

        There can be :py:class:`~.command.FormatterCommand` instance instead of
        a row. See documentation of this class for list of allowed commands.

        :param rows: List of rows to print.
        :type rows: list or generator
        """
        super(TableFormatter, self).produce_output(rows)
        self.print_stash()

class CsvFormatter(ListFormatter):
    """
    Renders data in a csv (Comma-separated values) format.

    This formatter supports following commands:
        * :py:class:`~.command.NewHostCommand`
        * :py:class:`~.command.NewTableCommand`
        * :py:class:`~.command.NewTableHeaderCommand`
    """

    def render_value(self, val):
        if isinstance(val, basestring):
            if isinstance(val, unicode):
                val.encode(self.encoding)
            val = '"%s"' % val.replace('"', '""')
        elif val is None:
            val = ''
        else:
            val = str(val)
        return val

    def print_text_row(self, row):
        self.print_line(",".join(self.render_value(v) for v in row))
        self.line_counter += 1

class SingleFormatter(Formatter):
    """
    Meant to render and print attributes of single object. Attributes are
    rendered as a list of assignments of values to variables (attribute names).

    This formatter supports following commands:
        * :py:class:`~.command.NewHostCommand`
    """

    def produce_output(self, data):
        """
        Render and print attributes of single item.

        There can be a :py:class:`~.command.FormatterCommand` instance instead
        of a data. See documentation of this class for list of allowed
        commands.

        :param data: Is either a pair of property names with list of values or
            a dictionary with property names as keys. Using the pair allows to
            order the data the way it should be printing. In the latter case
            the properties will be sorted by the property names.
        :type data: tuple or dict
        """
        if isinstance(data, fcmd.NewHostCommand):
            self.print_host(data.hostname)
            return

        if not isinstance(data, (tuple, dict)):
            raise ValueError("data must be tuple or dict")

        if isinstance(data, tuple):
            if not len(data) == 2:
                raise ValueError(
                    "data must contain: (list of columns, list of rows)")
            dataiter = itertools.izip(data[0], data[1])
        else:
            dataiter = itertools.imap(
                    lambda k: (k, self.render_value(data[k])),
                    sorted(data.keys()))
        for name, value in dataiter:
            self.print_line("{0}={1}", name, value)
            self.line_counter += 1

class ShellFormatter(SingleFormatter):
    """
    Specialization of :py:class:`SingleFormatter` having its output executable
    as a shell script.

    This formatter supports following commands:
        * :py:class:`~.command.NewHostCommand`
    """

    def render_value(self, val):
        if isinstance(val, basestring):
            if isinstance(val, unicode):
                val.encode(self.encoding)
            val = "'%s'" % val.replace("'", "\\'")
        elif val is None:
            val = ''
        else:
            val = str(val)
        return val

class ErrorFormatter(ListFormatter):
    """
    Render error strings for particular host. Supported commands:
        * :py:class:`~.command.NewHostCommand`
    """
    def __init__(self, stream, padding=4):
        super(ErrorFormatter, self).__init__(stream, padding)

    def print_row(self, data):
        if isinstance(data, Exception):
            if isinstance(data, errors.CIMError):
                self.print_text_row("%s: %s" % (data.args[1], data.message))
            elif not isinstance(data, errors.LmiFailed):
                self.print_text_row("(%s) %s" % (data.__class__.__name__,
                    str(data)))
            else:
                self.print_text_row(data)
        else:
            self.print_text_row(data)

    def print_host(self, hostname):
        self.out.write("host %s\n" % hostname)
        self.host_counter += 1
        self.table_counter = 0
        self.line_counter = 0

    def produce_output(self, rows):
        for row in rows:
            if isinstance(row, fcmd.NewHostCommand):
                self.print_host(row.hostname)
            elif isinstance(row, fcmd.NewTableCommand):
                self.print_table_title(row.title)
            else:
                self.print_row(row)




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
This module defines LmiCheckResult command class and related utilities.
"""

import abc

from lmi.scripts.common import Configuration
from lmi.scripts.common import get_logger
from lmi.scripts.common import formatter
from lmi.scripts.common import errors
from lmi.scripts.common.command import meta
from lmi.scripts.common.command.session import LmiSessionCommand

LOG = get_logger(__name__)

class LmiResultFailed(errors.LmiFailed):
    """
    Exception raised when associated function returns unexpected result. This
    is evaluated by :py:meth:`LmiCheckResult.check_result` method.
    """
    pass

def _make_result_failed(expected, result):
    """
    Instantiate :py:exc:`LmiResultFailed` exception with descriptive message
    composed of what was expected and what was returned instead.

    :rtype: :py:class:`LmiResultFailed`
    """
    return LmiResultFailed('failed (%s != %s)' % (repr(expected), repr(result)))

class LmiCheckResult(LmiSessionCommand):
    """
    Run an associated action and check its result. It implicitly makes no
    output if the invocation is successful and expected result matches.

    List of additional recognized properties:

        ``EXPECT`` :
            Value, that is expected to be returned by invoked associated
            function. This can also be a callable taking two arguments:

                1. options - Dictionary with parsed command line options
                   returned by ``docopt``.
                2. result - Return value of associated function.

    Using metaclass: :py:class:`~.meta.CheckResultMetaClass`.
    """
    __metaclass__ = meta.CheckResultMetaClass

    def __init__(self, *args, **kwargs):
        LmiSessionCommand.__init__(self, *args, **kwargs)

    def formatter_factory(self):
        return formatter.TableFormatter

    @abc.abstractmethod
    def check_result(self, options, result):
        """
        Check the returned value of associated function.

        :param dictionary options: Dictionary as returned by ``docopt`` parser
            after running
            :py:meth:`~.endpoint.LmiEndPointCommand.transform_options`.
        :param result: Any return value that will be compared against what is
            expected.
        :returns:  Whether the result is expected value or not. If ``tuple``
            is returned, it contains ``(passed_flag, error_description)``.
        :rtype: boolean or tuple.
        """
        raise NotImplementedError("check_result must be overriden in subclass")

    def take_action(self, connection, args, kwargs):
        """
        Invoke associated method and check its return value for single host.

        :param list args: List of arguments to pass to the associated
            function.
        :param dictionary kwargs: Keyword arguments to pass to the associated
            function.
        :returns: Exit code (0 on success).
        :rtype: integer
        """
        try:
            result = self.execute_on_connection(connection, *args, **kwargs)
            passed = self.check_result(self._options, result)
            if isinstance(passed, tuple):
                if len(passed) != 2:
                    raise errors.LmiUnexpectedResult('check_result() must'
                        ' return either boolean or (passed_flag,'
                        ' error_description), not "%s"' % repr(passed))
                if not passed[0]:
                    raise LmiResultFailed(passed[1])
            elif not passed and hasattr(self.check_result, 'expected'):
                err = _make_result_failed(self.check_result.expected, result)
                raise err
        except LmiResultFailed:
            raise
        except Exception as err:
            LOG().debug("Failed to execute wrapped function.", exc_info=err)
            raise
        return 0

    def process_host_result(self, hostname, success, result):
        pass

    def process_session_results(self, session, results):
        if len(self.session) > 1:
            LOG().debug('Successful runs: %d\n',
                    len([r for r in results.values() if r[0]]))
        LmiSessionCommand.process_session_results(self, session, results)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import hashlib
import logging
import urlparse

from lmi.shell import wbem

from lmi.shell.LMIObjectFactory import LMIObjectFactory

LOCALHOST_VARIANTS = (
    "localhost",
    "localhost.localdomain",
    "localhost4",
    "localhost4.localdomain4",
    "localhost6",
    "localhost6.localdomain6",
    "127.0.0.1",
    "::1",
)


# By default, the LMIShell does not use exceptions; LMIReturnValue
# is used instead (with proper error string).
class LMIUseExceptionsHelper(object):
    """
    Singleton helper class used for storing a bool flag, which defines,
    if the LMIShell should propagate exceptions or dump them.
    """
    _instance = None

    def __new__(cls):
        """
        Return a new :py:class:`.LMIUseExceptionsHelper` instance, if no shared
        object is present; otherwise an existing instance is returned. By
        default the ``use_exceptions`` flag is set to False.
        """
        if cls._instance is None:
            cls._instance = super(LMIUseExceptionsHelper, cls).__new__(cls)
            cls._instance._use_exceptions = False
        return cls._instance

    @property
    def use_exceptions(self):
        """
        :returns: whether the LMIShell should propagate the exceptions, or
            throw them away
        :rtype: bool
        """
        return self._use_exceptions

    @use_exceptions.setter
    def use_exceptions(self, use=True):
        """
        Property setter, which modifies the bool flag, which indicates, if the
        LMIShell should propagate the exceptions, or dump them.

        :param bool use: specifies, whether to use exceptions in LMIShell
        """
        self._use_exceptions = use


class LMIPassByRef(object):
    """
    Helper class used for passing a value by reference. It uses the advantage
    of python, where all the dictionaries are passed by reference.

    :param val: value, which will be passed by reference

    Example of usage:

    .. code-block:: python

        by_ref = LMIPassByRef(some_value)
        by_ref.value == some_value
    """
    def __init__(self, val):
        self._val = {0: val}

    @property
    def value(self):
        """
        :returns: value passed by reference.
        """
        return self._val[0]

    @value.setter
    def value(self, new_val):
        """
        Property setter for the value passed by reference.

        :param new_val: new value, which is passed by reference
        """
        self._val[0] = new_val


def lmi_get_use_exceptions():
    """
    :returns: whether the LMIShell should use the exceptions, or throw them
        away
    :rtype: bool
    """
    return LMIUseExceptionsHelper().use_exceptions


def lmi_set_use_exceptions(use=True):
    """
    Sets a global flag indicating, if the LMIShell should use the exceptions,
    or throw them away.

    :param bool use: specifies, whether the LMIShell should use the exceptions
    """
    LMIUseExceptionsHelper().use_exceptions = use


def lmi_raise_or_dump_exception(e=None):
    """
    Function which either raises an exception, or throws it away.

    :param Exception e: exception, which will be either raised or thrown away
    """
    if not lmi_get_use_exceptions():
        return
    et, ei, tb = sys.exc_info()
    if e is None:
        raise et, ei, tb
    else:
        raise type(e), e, tb


def lmi_parse_uri(uri):
    """
    Parses URI into scheme, hostname, port, username and password.
    """
    scheme, netloc, path = urlparse.urlparse(uri)[0:3]
    hp = netloc
    if "@" in netloc:
        up, hp = netloc.split("@")
        try:
            username, password = up.split(":")
        except ValueError:
            raise ValueError("Wrong uri format")
    else:
        username = None
        password = None

    if ":" in hp:
        hostname, str_port = hp.split(":")
        port = int(str_port)
    else:
        hostname = hp
        if scheme == "http":
            port = 5985
        else:
            port = 5986

    return scheme, hostname, port, path, username, password


def _lmi_do_cast(t, value, cast):
    """
    Helper function, which preforms the actual cast.

    :param string t: string of CIM type
    :param value: variable to cast
    :param dictionary cast: dictionary with :samp:`type : cast_func`
    :returns: cast value
    """
    cast_func = cast.get(t.lower(), lambda x: x)
    if isinstance(value, (dict, wbem.NocaseDict)):
        return wbem.NocaseDict(
            dict((k, _lmi_do_cast(t, v, cast)) for k, v in value.iteritems()))
    elif isinstance(value, list):
        return [_lmi_do_cast(t, v, cast) for v in value]
    elif isinstance(value, tuple):
        return (_lmi_do_cast(t, v, cast) for v in value)
    return cast_func(value) if value is not None else value


def lmi_cast_to_cim(t, value):
    """
    Casts the value to CIM type.

    :param string t: string of CIM type
    :param value: variable to cast
    :returns: cast value in :py:mod:`wbem` type
    """
    cast = {
        "sint8": lambda x: wbem.Sint8(x),
        "uint8": lambda x: wbem.Uint8(x),
        "sint16": lambda x: wbem.Sint16(x),
        "uint16": lambda x: wbem.Uint16(x),
        "sint32": lambda x: wbem.Sint32(x),
        "uint32": lambda x: wbem.Uint32(x),
        "sint64": lambda x: wbem.Sint64(x),
        "uint64": lambda x: wbem.Uint64(x),
        "string": lambda x: unicode(x, "utf-8") if isinstance(x, str) else x,
        "reference": lambda x: lmi_instance_to_path(x)
    }
    return _lmi_do_cast(t, value, cast)


def lmi_cast_to_lmi(t, value):
    """
    Casts the value to LMI (python) type.

    :param string t: string of CIM type
    :param value: variable to cast
    :returns: cast value in :py:mod:`python` native type
    """
    cast = {
        "sint8": lambda x: int(x),
        "uint8": lambda x: int(x),
        "sint16": lambda x: int(x),
        "uint16": lambda x: int(x),
        "sint32": lambda x: int(x),
        "uint32": lambda x: int(x),
        "sint64": lambda x: int(x),
        "uint64": lambda x: int(x),
    }
    return _lmi_do_cast(t, value, cast)


def lmi_wrap_cim_namespace(conn, cim_namespace_name):
    """
    Helper function, which returns wrapped CIM namespace in
    :py:class:`.LMINamespace`.

    :param LMIConnection conn: connection object
    :param string cim_namespace_name: CIM namespace name
    :returns: wrapped CIM namespace into :py:class:`.LMINamespace`
    """
    return LMIObjectFactory().LMINamespace(conn, cim_namespace_name)


def lmi_wrap_cim_class(conn, cim_class_name, cim_namespace_name):
    """
    Helper function, which returns wrapped :py:class:`wbem.CIMClass` into
    :py:class:`.LMIClass`.

    :param LMIConnection conn: connection object
    :param string cim_class_name: string containing :py:class:`wbem.CIMClass`
        name
    :param string cim_namespace_name: string containing
        :py:class:`wbem.CIMNamespace` name, or None, if the namespace is not
        known
    :returns: wrapped :py:class:`wbem.CIMClass` into :py:class:`.LMIClass`
    """
    lmi_namespace = None
    if cim_namespace_name:
        lmi_namespace = lmi_wrap_cim_namespace(conn, cim_namespace_name)
    return LMIObjectFactory().LMIClass(conn, lmi_namespace, cim_class_name)


def lmi_wrap_cim_instance(conn, cim_instance, cim_class_name,
                          cim_namespace_name):
    """
    Helper function, which returns wrapped :py:class:`wbem.CIMInstance` into
    :py:class:`.LMIInstance`.

    :param LMIConnection conn: connection object
    :param CIMInstance cim_instance: :py:class:`wbem.CIMInstance` object to be
        wrapped
    :param string cim_class_name: :py:class:`wbem.CIMClass` name
    :param string cim_namespace_name: :py:class:`wbem.CIMNamespace` name, or
        None, if the namespace is not known
    :returns: wrapped :py:class:`wbem.CIMInstance` into
        :py:class:`.LMIInstance`
    """
    lmi_class = lmi_wrap_cim_class(conn, cim_class_name, cim_namespace_name)
    return LMIObjectFactory().LMIInstance(conn, lmi_class, cim_instance)


def lmi_wrap_cim_instance_name(conn, cim_instance_name):
    """
    Helper function, which returns wrapped :py:class:`wbem.CIMInstanceName`
    into :py:class:`.LMIInstanceName`.

    :param LMIConnection conn: connection object
    :param CIMInstanceName cim_instance_name: :py:class:`wbem.CIMInstanceName`
        object to be wrapped
    :returns: wrapped :py:class:`wbem.CIMInstanceName` into
        :py:class:`.LMIInstanceName`
    """
    return LMIObjectFactory().LMIInstanceName(conn, cim_instance_name)


def lmi_wrap_cim_method(conn, cim_method_name, lmi_instance):
    """
    Helper function, which returns wrapped :py:class:`wbem.CIMMethod` into
    :py:class:`.LMIMethod`.

    :param LMIConnection conn: connection object
    :param string cim_method_name: method name
    :param LMIInstance lmi_instance: object, on which the method call will be
        issued
    :returns: wrapped :py:class:`wbem.CIMMethod` into :py:class:`.LMIMethod`
    """
    return LMIObjectFactory().LMIMethod(conn, lmi_instance, cim_method_name)


def lmi_transform_to_lmi(conn, value):
    """
    Transforms returned values from a method call into LMI wrapped objects.
    Returns transformed input, where :py:class:`wbem.CIMInstance` and
    :py:class:`wbem.CIMInstanceName` are wrapped into LMI wrapper classes and
    primitive types are cast to python native types.

    :param LMIConnection conn: connection object
    :param value: object to be transformed into :py:mod:`python` type from
        :mod:`wbem` one
    :returns: transformed py::mod:`wbem` object into LMIShell one
    """
    if isinstance(value, wbem.CIMInstance):
        namespace = value.path.namespace if value.path else None
        return lmi_wrap_cim_instance(conn, value, value.classname, namespace)
    elif isinstance(value, wbem.CIMInstanceName):
        return lmi_wrap_cim_instance_name(conn, value)
    elif isinstance(value, wbem.CIMInt):
        return int(value)
    elif isinstance(value, wbem.CIMFloat):
        return float(value)
    elif isinstance(value, (dict, wbem.NocaseDict)):
        return wbem.NocaseDict(
            dict(
                (k, lmi_transform_to_lmi(conn, v))
                for k, v in value.iteritems()))
    elif isinstance(value, list):
        return [lmi_transform_to_lmi(conn, val) for val in value]
    elif isinstance(value, tuple):
        return (lmi_transform_to_lmi(conn, val) for val in value)
    return value


def lmi_transform_to_cim_param(t, value):
    """
    Helper function for method calls, which transforms input object into
    :py:class:`wbem.CIMInstanceName` object. Members if lists, dictionaries and
    tuples are transformed as well. The function does not cast numeric types.

    :param string t: string of CIM type
    :param value: object to be transformed to :py:mod:`wbem` type.
    :returns: transformed LMIShell's object into :py:mod:`wbem` one
    """
    if isinstance(value, LMIObjectFactory().LMIInstance):
        return value.wrapped_object.path
    elif isinstance(value, LMIObjectFactory().LMIInstanceName):
        return value.wrapped_object
    elif isinstance(value, (dict, wbem.NocaseDict)):
        return wbem.NocaseDict(
            dict(
                (k, lmi_transform_to_cim_param(t, val))
                for k, val in value.iteritems()))
    elif isinstance(value, list):
        return [lmi_transform_to_cim_param(t, val) for val in value]
    elif isinstance(value, tuple):
        return (lmi_transform_to_cim_param(t, val) for val in value)
    return lmi_cast_to_cim(t, value)


def lmi_isinstance(lmi_obj, lmi_class):
    """
    Function returns True if :samp:`lmi_obj` is an instance of a
    :samp:`lmi_class`, False otherwise. When passed :py:class:`.LMIInstance`,
    :py:class:`.LMIInstanceName` as :samp:`lmi_obj` and :samp:`lmi_class` is of
    :py:class:`.LMIClass` type, function can tell, if such :samp:`lmi_obj` is
    direct instance of :py:class:`.LMIClass`, or it's super class.

    If :samp:`lmi_obj` and :samp:`lmi_class` is not instance of mentioned
    classes, an exception will be raised.

    :param lmi_obj: instance of :py:class:`.LMIInstance` or
        :py:class:`.LMIInstanceName` which is checked, if such instance is
        instance of the ``lmi_class``
    :param LMIClass lmi_class: instance of :py:class:`.LMIClass` object
    :returns: whether **lmi_obj** is instance of **lmi_class**
    :rtype: bool
    :raises: :py:exc:`TypeError`
    """
    cls = (
        LMIObjectFactory().LMIInstance,
        LMIObjectFactory().LMIInstanceName)
    if not isinstance(lmi_obj, cls) or \
            not isinstance(lmi_class, LMIObjectFactory().LMIClass):
        errorstr = "Use with types LMIInstance/LMIInstanceName and LMIClass"
        lmi_raise_or_dump_exception(TypeError(errorstr))
        return False
    client = lmi_obj._conn.client
    classname = lmi_obj.classname
    namespace = lmi_obj.namespace
    while classname:
        if classname == lmi_class.classname:
            return True
        classname, _, errorstr = client.get_superclass(classname, namespace)
    return False


def lmi_associators(assoc_classes):
    """
    Helper function to speed up associator traversal. Returns a list of tuples,
    where each tuple contains :py:class:`.LMIInstance` objects, which are in
    association.

    :param list assoc_classes: list of :py:class:`.LMIClass` objects, for which
        the associations will be returned
    :returns: list of tuples of :py:class:`.LMIInstance` objects in association
    """
    def make_key(path):
        path.host = None
        sum = hashlib.md5(
            path.classname.lower() + path.namespace.lower() +
            str(dict((k.lower(), v) for k, v in path.keybindings.iteritems())))
        return sum.hexdigest()

    result = []

    instances = {}
    for assoc_class in assoc_classes:
        conn = assoc_class._conn
        assoc_class.fetch()

        # Reference properties
        ref_props = [
            pname
            for pname, prop in assoc_class._cim_class.properties.iteritems()
            if prop.type == "reference"
        ]

        # Reference class names
        ref_class_names = [
            assoc_class._cim_class.properties[ref_prop].reference_class
            for ref_prop in ref_props
        ]

        # Get instances, which will be joined as associators
        for ref_class_name in ref_class_names:
            inst_list, out, err = conn.client.get_instances(ref_class_name)
            instances.update(
                dict((make_key(inst.path), inst) for inst in inst_list))

        # Join associated objects
        assoc_instance_names, out, err = conn.client.get_instance_names(
            assoc_class.classname)
        for assoc in assoc_instance_names:
            ref_list = []
            for ref_prop in ref_props:
                path = assoc[ref_prop]
                # XXX: Some association classes report in key property a class,
                # which its instances do not refer to.
                inst = instances.get(make_key(path), None)
                if inst:
                    ref_list.append(
                        lmi_wrap_cim_instance(
                            conn, inst, inst.classname, inst.path.namespace))

            if len(ref_list) == len(ref_class_names):
                result.append(tuple(ref_list))

    return result


def lmi_instance_to_path(instance):
    """
    Helper function, which returns :py:class:`wbem.CIMInstanceName` extracted
    out of input instance.

    :param instance: object, which can be instance of following classes:

        * :py:class:`wbem.CIMInstance`
        * :py:class:`wbem.CIMInstanceName`
        * :py:class:`.LMIInstance`
        * :py:class:`.LMIInstanceName`

    :returns: extracteed :py:class:`wbem.CIMInstanceName` object
    :raises: :py:exc:`TypeError`
    """
    if isinstance(instance, wbem.CIMInstance):
        return instance.path
    elif isinstance(instance, wbem.CIMInstanceName):
        return instance
    elif isinstance(instance, LMIObjectFactory().LMIInstance):
        return instance.wrapped_object.path
    elif isinstance(instance, LMIObjectFactory().LMIInstanceName):
        return instance.wrapped_object
    raise TypeError(
        "instance has to be wbem.CIMInstance(Name) or LMIInstance(Name)")

def lmi_is_localhost(uri):
    """
    Helper function, which returns True, if URI points to localhost.

    :param str uri: URI to check
    """
    return uri in LOCALHOST_VARIANTS



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Utility functions for command inspection.
"""
from lmi.scripts.common import errors
from lmi.scripts.common.command import LmiBaseCommand
from lmi.scripts.common.command import LmiCommandMultiplexer

def get_subcommand_names(command):
    """
    :param command: Either a multiplexer command or top-level one.
    :returns: Names of children commands.
    :rtype: list
    """
    if not isinstance(command, LmiBaseCommand):
        raise TypeError("command must be an instance of LmiBaseCommand")
    if isinstance(command, LmiCommandMultiplexer):
        return command.child_commands().keys()
    if command.parent is None:  # top level command
        return command.app.command_manager.command_names
    raise ValueError("command must be either multiplexer or top-level command")

def get_subcommand_factory(command, name):
    """
    :param command: Either a multiplexer command or top-level one.
    :returns: Callable returning an instance of
        :py:class:`~lmi.scripts.common.command.multiplexer.LmiCommandMultiplexer`
    :rtype: callable
    """
    if isinstance(command, LmiCommandMultiplexer):
        try:
            return command.child_commands()[name]
        except KeyError:
            cmd_path = command.cmd_name_parts
            cmd_path.append(name)
            raise errors.LmiCommandNotFound(" ".join(cmd_path))
    if command.parent is None:  # top level command
        return command.app.command_manager.find_command(name)
    raise ValueError("command must be either multiplexer or top-level command")

